# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import http.server
import hashlib
import logging
import os
import secrets
import simplejson as json
import socket
import time
import re
import urllib.request, urllib.parse, urllib.error

try:
    import pam #apt-get install python-pampy or pacman -S python-pam on ArchLinuxARM
except:
    pam = None

import config
import joystick_interface
import log
import paths
import platforms
import rights
import user_login
import version


class WebInterfaceHandler(http.server.BaseHTTPRequestHandler):

    WRITE_CHUNK_SIZE = 16384
    URL = config.get_settings()['URL']
    COOKIE_TOKEN_LENGHT = 256

    @staticmethod
    def decode_url_to_unicode(url):
        return urllib.parse.unquote(url.decode('utf-8').replace('+', '%20'))

    @staticmethod
    def htmlify_text(text, ignore_endofline=False):
        text = text.replace('\t', '&emsp;')
        if not ignore_endofline:
            text = text.replace('\n', '<br />')
        return text

    def log_message(self, format, *args): # otherwise every request will be logged
        return

    def setup(self):
        self.logger = logging.getLogger(__name__)
        self.request.settimeout(12)
        http.server.BaseHTTPRequestHandler.setup(self)
        self.page_former = self.server.page_former
        self.host, _ = self.client_address[:2]
        try:
            # IP of server's socket that received the connection
            self.connection_ip = str(self.connection.getsockname()[0])
        except:
            # fallback if OS refused to give us the real IP
            self.connection_ip = self.host

    def write_with_autoreplace(self, page, response=200, headers = {}):
        version_string = "v.%s.%s_%s" % (version.version, version.build, version.branch)
        is_text_page = type(page) == str
        if is_text_page: #binary data like images will have type bytes and doesn't need patching
            page = page.replace('!!!VERSION!!!', version_string)
            page = page.replace('!!!URL!!!', self.URL.replace('cli-cloud', 'cloud'))
            page = page.replace('!!!HTTPS_WARNING!!!', self.form_http_warning())
        if self.server.app.stop_flag:
            headers["Connection"] = "close"
        else:
            headers["Connection"] = "keep-alive"
        if is_text_page: # vital for correct length calculation
            length = len(page.encode("utf-8"))
        else:
            length = len(page)
        headers["Content-Length"] = length
        if not "Content-Type" in headers:
            headers["Content-Type"] = "text/html; charset=utf-8"
        try:
            self.send_response(response)
        except (ConnectionResetError, BrokenPipeError):
            self.logger.warning("Lost connection to browser")
        except Exception:
            self.logger.exception('Error while sending response on %s: ' % self.path)
        try:
            for keyword, value in headers.items():
                self.send_header(keyword, value)
            self.end_headers()
        except (ConnectionResetError, BrokenPipeError):
            self.logger.warning("Lost connection to browser")
        except Exception:
            self.logger.exception('Error while sending headers %s: ' % self.path)
        else:
            if is_text_page:
                page = page.encode('utf-8')
            chunk_start = 0
            while chunk_start < length:
                chunk_end = max(chunk_start + self.WRITE_CHUNK_SIZE, length)
                try:
                    self.wfile.write(page[chunk_start:chunk_end])
                except (ConnectionError, socket.error) as e:
                    self.logger.warning("Lost connection to browser. Error:%s" % str(e))
                    break
                except Exception:
                    self.logger.exception('Error while writing page to %s: ' % self.path)
                    break
                chunk_start = chunk_end

    def form_http_warning(self):
        message = ''
        if not self.server.HTTPS:
            https_port = config.get_settings()['web_interface']['port_https']
            if https_port:
                message = '<div id="http_secure_message">For security reasons we strongly recommend to use HTTPS url to access the cloud client interface.' \
                        ' <a href="https://%s/">Click here to switch to HTTPS</a>.</div>' % self.connection_ip
        return message

    def parse_post(self, *value_names):
        posts = []
        if "multipart/form-data" in self.headers.get('Content-Type', ''):
            message = "Error: method parse_post is incapable of parsing multipart posts"
            self.logger.error(message)
        else:
            content_length = int(self.headers.get('Content-Length'))
            body = self.rfile.read(content_length)
            body = self.decode_url_to_unicode(body)
            self.logger.debug('Path: %s\nPost: %s\nContent-Length: %s\n' % (self.path, body, content_length))
            post_dict = urllib.parse.parse_qs(body)
            for name in value_names:
                if name in post_dict:
                    posts.append(post_dict[name][0])
        if len(value_names) == 1:
            return posts[0]
        if len(posts) < len(value_names):
            posts.extend([''] * (len(value_names) - len(posts)))
        self.logger.debug("Post parsed as: %s" % posts)
        return posts

    def get_cookie(self, key):
        cookies = self.headers.get('Cookie', None)
        if cookies:
            cookies = cookies.split('; ')
            for cookie in cookies:
                if cookie.startswith(key):
                    return cookie.split('%s=' % key)[-1]

    def fix_cookie_name_https(self, key):
        if self.server.HTTPS:
            return key + '-https'
        return key

    def write_message(self, message, show_time=2, response=200, headers={},\
            redirect="/", button_text="", button_url="/"):
        page = self.server.web_content_files['templates']['pages']['message.html']
        page = page.replace('!!!MESSAGE!!!', message)
        if show_time:
            page = page.replace('!!!SHOW_TIME!!!', str(show_time))
        else:
            page = page.replace('<meta http-equiv="refresh" content="!!!SHOW_TIME!!!; url=/" />', '')
        page = page.replace('url=/', 'url=' + redirect)
        if button_text:
            button_html = self.server.web_content_files['templates']['buttons']['generic_button.html']
            button_html = button_html.replace("%VALUE%", button_text)
            button_html = button_html.replace("%PATH%", button_url)
            button_html = button_html.replace("%METHOD%", "get")
            button_html = button_html.replace("%CLASS%", "big_button")
            page = page.replace("!!!BUTTON!!!", button_html)
        else:
            page = page.replace("!!!BUTTON!!!", "")
        self.write_with_autoreplace(page, response=response, headers=headers)

    def redirect(self, url):
        self.send_response(301)
        self.send_header('Location', url)
        self.end_headers()

    def path_contains(self, path):
        return path in self.path

    def need_remote_login(self):
        if self.host != config.LOCAL_IP:
            token = self.get_cookie(self.fix_cookie_name_https('token'))
            return token not in self.server.cookies

    @log.log_exception
    def do_GET(self):
        if self.path_contains('static'):
            self.get_static()
        elif self.need_remote_login():
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['remote_login.html'])
        elif self.path_contains('ajax'):
            self.ajax_do_GET()
        elif self.path_contains('user_login'):
            self.process_login()
        elif self.path_contains('logout'):
            self.process_logout()
        elif self.path_contains('quit'):
            self.quit_main_app()
        elif self.path_contains('show_logs'):
            self.show_logs()
        elif self.path_contains('show_release_notes'):
            filename = os.path.join(os.path.dirname(__file__), "..", "release_notes.txt")
            with open(filename) as f:
                content = f.read()
            self.wfile.write(content.encode("utf-8"))
        elif self.path_contains('clear_logs'):
            self.clear_logs()
        elif self.path_contains('open_logs'):
            log.open_settings_folder()
            self.write_message(self.page_former.form_back_button('/'), show_time = 0)
        elif self.path_contains('kill_conveyor'):
            self.kill_conveyor()
        elif self.path_contains('connect_to_network_printer'):
            self.write_with_autoreplace(self.page_former.form_connect_to_network_printer_page())
        elif self.path_contains('detect_network_clients'):
            page = self.page_former.form_detect_network_clients_page()
            if page:
                self.write_with_autoreplace(page)
            else:
                self.write_message('No 3DPrinterOS Clients detected in your network')
        elif self.path_contains('edit_settings'):
            self.write_with_autoreplace(self.page_former.form_edit_settings_page())
        elif self.path_contains('update_from_file'):
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['update_from_file.html'])
        elif self.path_contains('manage_settings'):
            self.write_with_autoreplace(self.page_former.form_manage_settings_page())
        elif self.path_contains('rpi_access_control'):
            if self.check_if_rpi():
                self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['rpi_access_control.html'])
        elif self.path_contains('telemetry'):
            self.telemetry()
        elif self.path_contains('plugins_page'):
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['plugins_page.html'])
        elif self.path_contains('check_host'):
            self.check_host()
        elif self.path_contains('check_makerware'):
            self.check_makerware()
        elif self.path_contains('check_linux_rights'):
            self.check_linux_rights()
        elif self.path_contains('check_if_printing'):
            self.check_if_printing()
        elif self.path_contains('telemetry'):
            self.telemetry()
        elif self.path_contains('network_printer_list_page'):
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['network_printer_list.html'])
        elif self.path_contains('network_printers_list'):
            self.network_printers_list()
        elif self.path_contains('plugins_list'):
            self.plugins_list()
        elif self.path_contains('install_plugin_page'):
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['install_plugin.html'])
        elif self.path_contains('upload_to_storage_page'):
            self.upload_to_storage_page()
        elif self.path_contains('enter_offline_mode'):
            self.server.app.set_offline_mode(True)
            self.write_message('Entering offline mode...', show_time=1)
        elif self.path_contains('exit_offline_mode'):
            self.server.app.set_offline_mode(False)
            self.write_message('Exiting offline mode...', show_time=1)
        elif self.path_contains('reload_html'):
            self.server.reload_web_content()
            self.write_message("OK", show_time=0.2)
        elif self.path_contains('detection_wizard_page'):
            self.get_detection_wizard_page()
        elif self.path_contains('detection_wizard_start'):
            self.detection_wizard_start()
        elif self.path_contains('/favicon.ico'):
            self.answer_with_image('favicon.ico')
        else:
            page = self.page_former.form_main_page()
            if page:
                self.server.app.detection_wizard.stop() # disable detection wizard if we go to main page
                self.write_with_autoreplace(page, headers={'Access-Control-Allow-Origin':'*'})
            else:
                self.write_message('Initializing...', show_time=1)
            # disable detection wizard when user is on main page

    def get_static(self):
        keys = self.path.split('/')[1:]
        web_content_files = dict(self.server.web_content_files)
        for key in keys:
            if key in web_content_files:
                if isinstance(web_content_files[key], dict):
                    web_content_files = web_content_files[key]
                elif isinstance(web_content_files[key], str):
                    if key.endswith('css'):
                        self.write_with_autoreplace(web_content_files[key], headers={'Content-Type': 'text/css'})
                    elif key.endswith('js'):
                        self.write_with_autoreplace(web_content_files[key], headers={'Content-Type': 'text/javascript'})
                    else:
                        self.write_with_autoreplace(web_content_files[key])
                elif isinstance(web_content_files[key], bytes):
                    if key.endswith('png'):
                        self.write_with_autoreplace(web_content_files[key], headers={'Content-Type': 'image/png'})
                    else:
                        self.logger.warning("Unknown web content binary file that is not png: %s" % key )
                else:
                    self.logger.warning('Unknown key while getting static files: ' + key)

    def ajax_do_GET(self):
        if self.path_contains('wizard'):
            self.ajax_wizard_do_GET()
        elif self.path_contains('get_printers_info'):
            self.write_with_autoreplace(self.page_former.get_printers_info(),
                                        headers={'Content-Type': 'text/html'})
        elif self.path_contains('get_updates_button'):
            if self.server.app and hasattr(self.server.app, 'updater') and self.server.app.updater.update_available:
                update_button_content = self.server.web_content_files['templates']['buttons']['get_updates_button.html']
            else:
                update_button_content = ''
            self.write_with_autoreplace(update_button_content, headers={'Content-Type': 'text/html'})
        elif self.path_contains('write_log_tail'):
            self.write_log_tail()
        elif self.path_contains('get_closing_client_info'):
            self.write_with_autoreplace(self.page_former.form_closing_message(), headers={'Content-Type': 'text/html'})
        elif self.path_contains('check_login'):
            login = bool(hasattr(self.server.app, "user_login") and self.server.app.user_login.login) or self.server.app.offline_mode
            response = json.dumps({'result': login})
            self.write_with_autoreplace(response, headers={'Content-Type':'application/json'})
        elif self.path_contains('is_offline'):
            response = json.dumps({'result': self.server.app.offline_mode})
            self.write_with_autoreplace(response, headers={'Content-Type':'application/json'})

    def ajax_wizard_do_GET(self):
        if self.path_contains('get_printers_info_without_controls'):
            self.write_with_autoreplace(self.page_former.get_printers_info(False, False, False, False))
        elif self.path_contains('get_printers_info_with_type_button'):
            self.write_with_autoreplace(self.page_former.get_printers_info(False, True, False, False))
        elif self.path_contains('get_printers_info_with_rename_button'):
            self.write_with_autoreplace(self.page_former.get_printers_info(True, False, False, False))
        elif self.path_contains('get_printers_info_with_workgroups_button'):
            self.write_with_autoreplace(self.page_former.get_printers_info(False, False, True, False))

    @log.log_exception
    def do_POST(self):
        if self.path_contains('remote_login'):
            self.process_remote_login()
        elif self.need_remote_login():
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['remote_login.html'])
        elif self.path_contains('ajax'):
            self.ajax_do_POST()
        elif self.path_contains('joystick'):
            self.joystick_do_POST()
        elif self.path_contains('/wizard'):
            self.wizard_do_POST()
        elif self.path_contains('login'):
            self.process_login()
        elif self.path_contains('quit_confirmation'):
            self.write_with_autoreplace(self.page_former.form_quit_confirm_page())
        elif self.path_contains('quit'):
            self.quit_main_app()
        elif self.path_contains('logout_confirmation'):
            self.write_with_autoreplace(self.page_former.form_logout_confirm_page())
        elif self.path_contains('logout'):
            self.process_logout()
        elif self.path_contains('kill_conveyor'):
            self.kill_conveyor()
        elif self.path_contains('add_user_groups'):
            self.add_user_groups()
        elif self.path_contains('ignore_groups_warning'):
            self.ignore_groups_warning()
        elif self.path_contains('get_updates'):
            self.get_updates()
        elif self.path_contains('update_software'):
            self.update_software()
        elif self.path_contains('choose_cam'):
            self.choose_cam()
        elif self.path_contains('switch_camera'):
            self.switch_camera()
        elif self.path_contains('choose_printer_type'):
            self.choose_printer_type()
        elif self.path_contains('confirm_printer_groups'):
            self.confirm_printer_groups()
        elif self.path_contains('report_problem'):
            self.report_problem()
        elif self.path_contains('send_problem_report'):
            self.send_problem_report()
        elif self.path_contains('restart_camera'):
            self.server.app.camera_controller.restart_camera()
            self.write_message('Camera module restarted', redirect="/manage_settings")
        elif self.path_contains('set_new_printer_name'):
            self.set_new_printer_name()
        elif self.path_contains('toggle_virtual_printer'):
            self.toggle_virtual_printer()
        elif self.path_contains('toggle_gpio'):
            self.toggle_gpio()
        elif self.path_contains('toggle_prt_in_dual_serial'):
            self.toggle_prt_in_dual_serial()
        elif self.path_contains('toggle_clisma'):
            self.toggle_clisma()
        elif self.path_contains('reset_printer_type'):
            self.reset_printer_type()
        elif self.path_contains('restore_default_settings'):
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['restore_default_settings.html'])
        elif self.path_contains('confirm_default_settings_restoring'):
            config.Config.instance().restore_default_settings()
            self.write_message('Default settings restored. Please restart app to apply changes.')
        elif self.path_contains('tray'):
            self.tray_do_POST()
        elif self.path_contains('process_network_printer'):
            self.process_network_printer()
        elif self.path_contains('confirm_detect_network_printers'):
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']
                                        ['confirm_detect_printers.html'])
        elif self.path_contains('detect_network_printers'):
            self.server.app.network_printers_detector.network_detector_run_flag = True
            self.write_message('<p>Detecting network printers...</p>'
                               'If you have a printer connected to your network<br />it will appear in printers list',
                               show_time=4)
        elif self.path_contains('process_pairing_network_printer'):
            self.do_GET()
        elif self.path_contains('forget_network_printer'):
            self.forget_network_printer()
        elif self.path_contains('save_settings'):
            self.save_settings()
        elif self.path_contains('settings'):
            self.settings_do_POST()
        elif self.path_contains('accept_update_from_file'):
            self.accept_update_from_file()
        elif self.path_contains('control_plugin'):
            self.control_plugin()
        elif self.path_contains('upload_plugin'):
            self.accept_and_install_plugin_file()
        elif self.path_contains('change_ssh_password'):
            self.change_ssh_password()
        elif self.path_contains('generate_ssh_keys'):
            self.generate_ssh_keys()
        elif self.path_contains('revoke_all_ssh_keys'):
            self.revoke_all_ssh_keys()
        elif self.path_contains('add_public_key'):
            self.add_public_key()
        elif self.path_contains('upload_to_storage'):
            self.upload_to_storage()
        elif self.path_contains('local_file_print_or_delete'):
            self.local_file_print_or_delete()
        elif self.path_contains('print_from_file_or_delete'):
            self.print_from_file_or_delete()
        elif self.path_contains('detection_wizard_action'):
            self.detection_wizard_action()
        elif self.path_contains('detection_wizard_form_action'):
            self.detection_wizard_form_action()
        elif self.path_contains('detection_wizard_start'):
            self.detection_wizard_start()
        else:
            self.write_message('Not found', 0, 404)

    def display_wizard_stage(self, stage_name):
        self.page_former.set_wizard_stage(stage_name)
        self.do_GET()

    def settings_do_POST(self):
        settings = config.get_settings()
        if self.path_contains('toggle_update'):
            settings['update']['enabled'] = not settings['update']['enabled']
            config.Config.instance().save_settings(settings)
            self.write_with_autoreplace(self.page_former.form_manage_settings_page())
        elif self.path_contains('toggle_auto_update'):
            settings['update']['auto_update_enabled'] = not settings['update']['auto_update_enabled']
            config.Config.instance().save_settings(settings)
            self.write_with_autoreplace(self.page_former.form_manage_settings_page())
        elif self.path_contains('set_update_check_pause'):
            try:
                check_pause = int(self.parse_post('check_pause'))
            except:
                self.write_message('Setting update check pause failed: wrong value!')
            else:
                settings['update']['check_pause'] = int(check_pause)
                config.Config.instance().save_settings(settings)
                self.write_with_autoreplace(self.page_former.form_manage_settings_page())

    def wizard_do_POST(self):
        if self.path_contains('greetings'):
            self.page_former.wizard_enabled = True
            self.display_wizard_stage('greetings')
        elif self.path_contains('close_wizard'):
            self.close_wizard()
        elif self.path_contains('connect_printer'):
            self.display_wizard_stage('connect_printer')
        elif self.path_contains('choose_printer_type'):
            self.display_wizard_stage('choose_printer_type')
        elif self.path_contains('choose_live_view_mode'):
            self.display_wizard_stage('choose_live_view_mode')
        elif self.path_contains('tray_icon'):
            self.display_wizard_stage('tray_icon')
        elif self.path_contains('more_features'):
            self.display_wizard_stage('more_features')
        elif self.path_contains('rename_printer'):
            self.display_wizard_stage('rename_printer')
        elif self.path_contains('printer_workgroups'):
            self.display_wizard_stage('printer_workgroups')
        elif self.path_contains('restore_default_settings'):
            self.display_wizard_stage('restore_default_settings')
        elif self.path_contains('report_problem'):
            self.display_wizard_stage('report_problem')
        elif self.path_contains('final'):
            self.display_wizard_stage('final')
        else:
            self.write_message('Wizard stage not found', 0, 404)

    def ajax_do_POST(self):
        if self.path_contains('rename_printer'):
            self.rename_printer()
        elif self.path_contains('select_printer_groups'):
            self.select_printer_groups()
        elif self.path_contains('select_printer_type'):
            self.select_printer_type()
        elif self.path_contains('manage_printer'):
            self.manage_printer()
        elif self.path_contains('request_printer_type_reset'):
            self.request_printer_type_reset()
        elif self.path_contains('toggle_local_mode'):
            self.toggle_local_mode()
        elif self.path_contains('cancel_printing'):
            self.cancel_print()
        else:
            self.write_message('Not found', 0, 404)

    def tray_do_POST(self):
        response = None
        if self.path_contains('get_active_printers'):
            response = self.page_former.form_json_active_printers()
        elif self.path_contains('get_not_ready_printers'):
            response = self.page_former.form_json_active_printers(include_ready=False)
        elif self.path_contains('confirm_identical_printers'):
            response = self.page_former.form_json_confirm_identical_printers()
        elif self.path_contains('updates_available'):
            response = self.page_former.form_json_updates_available()
        if response:
            self.write_with_autoreplace(response.encode("utf-8"))

    def toggle_virtual_printer(self):
        self.server.app.toggle_virtual_printer()
        if self.page_former.wizard_enabled:
            self.display_wizard_stage('connect_printer')
        else:
            self.write_with_autoreplace(self.page_former.form_manage_settings_page())

    def toggle_gpio(self):
        self.server.app.gpio_interface.toggle()
        self.write_with_autoreplace(self.page_former.form_manage_settings_page())

    def toggle_prt_in_dual_serial(self):
        settings = config.get_settings()
        settings['backward_compatible_serials'] = not settings['backward_compatible_serials']
        config.Config.instance().save_settings(settings)
        self.write_with_autoreplace(self.page_former.form_manage_settings_page())

    def toggle_clisma(self):
        rights.toggle_clisma()
        self.write_with_autoreplace(self.page_former.form_manage_settings_page())

    def rename_printer(self):
        usb_info = self.parse_post('usb_info')
        page = self.page_former.form_rename_printer_page(usb_info)
        if page:
            self.write_with_autoreplace(page)
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def set_new_printer_name(self):
        printer_name, usb_info = self.parse_post('printer_name', 'usb_info')
        printer_interface = self.server.find_printer_interface(usb_info)
        if printer_interface:
            if printer_name:
                printer_interface.request_printer_rename(printer_name)
                self.write_message('Printer rename requested: "' + printer_name +'"')
            else:
                self.write_message('Wrong printer name')
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def select_printer_groups(self):
        usb_info = self.parse_post('usb_info')
        page = self.page_former.form_printer_groups_page(usb_info)
        if page:
            self.write_with_autoreplace(page)
        else:
            if config.get_settings()['protocol']['encryption']:
                prefix = 'https://'
            else:
                prefix = 'http://'
            back_button = self.page_former.form_back_button('/')
            self.write_message("<p>To create a new Workgroup Access code, please check Admin tab \
                                at <a href='" + prefix + \
                               "!!!URL!!!/workgroup' target='_blank'>!!!URL!!!/workgroup</a></p>\
                                Workgroup access codes is a premium \
                                feature, to use that please contact our sales: \
                                <a href='mailto:info@3dprinteros.com'>info@3dprinteros.com</a>" \
                               + back_button , show_time=0)

    def confirm_printer_groups(self):
        usb_info = self.get_unicode_string_from_current_path(b'/confirm_printer_groups/')
        printer_interface = self.server.find_printer_interface(usb_info)
        if printer_interface:
            printer_groups = printer_interface.get_groups()
            check_list = []
            for index in range(0, len(printer_groups)):
                check_list.append('selected_group_id_%d' % index)
            selected_groups_id = self.parse_post(*check_list)
            if type(selected_groups_id) != list:
                selected_groups_id = [] if not selected_groups_id else [str(selected_groups_id)]
            for index, group in enumerate(printer_groups):
                printer_groups[index]['active'] = str(group['id']) in selected_groups_id
            printer_interface.request_printer_groups_selection(printer_groups)
            self.write_message('Groups for ' + printer_interface.printer_profile['name'] + ' successfully updated')
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def find_printer_type(self, printer_interface, type_name):
        types_to_select = getattr(printer_interface, "possible_printer_types", None)
        if types_to_select:
            for printer_type in types_to_select:
                if printer_type['name'] == type_name:
                    return printer_type

    def select_printer_type(self):
        usb_info = self.parse_post('usb_info')
        page = self.page_former.form_type_select_page(usb_info)
        if page:
            self.write_with_autoreplace(page)
        else:
            self.write_message('Error while displaying printer types')

    def choose_printer_type(self):
        printer_type, usb_info = self.parse_post('printer_type', 'usb_info')
        printer_interface = self.server.find_printer_interface(usb_info)
        if printer_interface:
            if printer_type:
                chosen_type = self.find_printer_type(printer_interface, printer_type)
                if chosen_type:
                    printer_interface.request_printer_type_selection(chosen_type)
                    message = 'Printer type selected:  ' + str(printer_type)
                else:
                    message = 'Error: printer type was not selected'
            else:
                message = 'Error: printer type was not selected'
        else:
            message = 'Printer not found. Probably it was disconnected.'
        self.write_message(message)

    def request_printer_type_reset(self):
        usb_info = self.parse_post('usb_info')
        page = self.page_former.form_type_reset_page(usb_info)
        if page:
            self.write_with_autoreplace(page)
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def get_unicode_string_from_current_path(self, substring_to_remove=b''):
        path = self.path.encode('utf-8').replace(substring_to_remove, b'')
        return self.decode_url_to_unicode(path)

    def accept_update_from_file(self):
        error = self.parse_and_save_multipart_file_upload(path=paths.UPDATE_FILE_PATH)
        if error:
            message = 'Failed to accept update package!\n' + error
            self.logger.warning(message)
            self.write_message(message)
        else:
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']
                                        ['update_from_file_accepted.html'])

    def process_remote_login(self):
        try:
            login, password = self.parse_post('login', 'password')
        except:
            self.write_message('Authorization failed! Wrong email or password.')
        else:
            if login and password and self.check_login_and_password(login, password):
                token = self.create_cookie_token()
                header_value = '%s=%s; HttpOnly' % (self.fix_cookie_name_https('token'), token)
                if self.server.HTTPS:
                    header_value += '; Secure'
                self.write_message('Login successful!<br><br>Processing...', headers= {'Set-cookie': header_value})
            else:
                self.write_message('Authorization failed! Wrong email or password.')

    def save_settings(self):
        settings = self.parse_post('settings')
        try:
            settings = json.loads(settings)
            config.Config.instance().save_settings(settings)
        except Exception as e:
            self.write_message('Error while saving settings<br>%s' % str(e))
        else:
            self.write_message('Settings successfully saved.<br>But will be applied only on after restart!', show_time=5)

    def joystick_do_POST(self):
        if self.path_contains('start'):
            self.start_joystick(self.parse_post('printer_id'))
        elif self.path_contains('send_command'):
            printer_id, command, value = self.parse_post('printer_id', 'command', 'value')
            joystick_interface.execute_command(self.server.find_printer_interface(printer_id), command, value)
            self.write_with_autoreplace('OK')
        elif self.path_contains('get_printer_info'):
            self.send_printer_info(self.parse_post('printer_id'))
        elif self.path_contains('send_gcode'):
            printer_id, gcode_string = self.parse_post('printer_id', 'gcode_string')
            joystick_interface.send_gcode(self.server.find_printer_interface(printer_id), gcode_string)
            self.write_with_autoreplace('OK')

    def send_printer_info(self, printer_id):
        printer_interface = self.server.find_printer_interface(printer_id)
        self.write_with_autoreplace(json.dumps(joystick_interface.get_printers_info(printer_interface)),
                                    headers={'Content-Type': 'application\json'})

    def start_joystick(self, printer_id):
        printer_interface = self.server.find_printer_interface(printer_id)
        if printer_interface and printer_interface.printer:
            if not printer_interface.local_mode:
                #self.write_with_autoreplace(self.page_former.form_local_mode_warning_page(printer_id))
                self.write_message("Please enable local mode for this printer in main screen", show_time=3)
            else:
                joystick_interface.init_callback(self.server.find_printer_interface(printer_id))
                self.write_with_autoreplace(self.page_former.form_joystick_page(printer_id))
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def process_network_printer(self):
        post = self.parse_post('printer_ip', 'printer_type')
        if len(post) == 2:
            printer_ip, printer_type = post[0], post[1]
            if printer_ip and printer_type:
                self.server.app.network_printers_detector.connect_to_network_printer(printer_ip, printer_type)
                self.write_message('Printer was added ' + printer_ip)
                return
        self.write_message('Printer IP and type must not be empty!')

    def toggle_local_mode(self):
        printer_interface = self.server.find_printer_interface(self.parse_post('usb_info'))
        if printer_interface:
            if printer_interface.local_mode:
                self.logger.info("Local mode off")
                printer_interface.turn_off_local_mode()
            else:
                self.logger.info("Local mode on")
                printer_interface.turn_on_local_mode()
            self.send_response(200)
            self.end_headers()
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def cancel_print(self):
        printer_id = self.parse_post('usb_info')
        if printer_id:
            try:
                printer_interface = self.server.find_printer_interface(printer_id)
                printer_interface.cancel_locally()
            except:
                self.logger.debug("Fail to cancel print job for %s" % printer_id)
        self.send_response(200)
        self.end_headers()

    def manage_printer(self):
        usb_info = self.parse_post('usb_info')
        page = self.page_former.form_manage_printer_page(usb_info)
        if page:
            self.write_with_autoreplace(page)
        else:
            self.write_message('Error while displaying printer managing page')

    def close_wizard(self, silent=False):
        settings = config.get_settings()
        settings['wizard_on_start'] = False
        config.Config.instance().save_settings(settings)
        self.page_former.wizard_enabled = False
        if not silent:
            self.write_message('Wizard successfully closed. Loading main page...')

    def reset_printer_type(self):
        usb_info = self.parse_post('usb_info')
        printer_interface = self.server.find_printer_interface(usb_info)
        if printer_interface:
            printer_interface.request_reset_printer_type()
            if printer_interface.offline_mode:
                self.write_message('Done', show_time=0.5)
            else:
                self.write_message('Sending printer type reset request...')
        else:
            self.write_message('Printer not found. Probably it was disconnected.')

    def choose_cam(self):
        page = self.page_former.form_choose_cam_page()
        if page:
            self.write_with_autoreplace(page)
        else:
            self.write_message('Live view feature disabled')

    def switch_camera(self):
        module_name = str(self.parse_post('module'))
        if module_name:
            self.server.app.camera_controller.switch_camera( \
                    new_camera_name=module_name, token=self.server.app.user_login.user_token)
            message = 'Live view mode switched to ' + module_name
        else:
            message = 'Live view mode not chosen'
        self.write_message(message, redirect="/manage_settings")

    def get_updates(self):
        self.write_with_autoreplace(self.page_former.form_get_updates_page())

    def update_software(self):
        result = self.server.app.updater.download_update()
        if result:
            self.write_message(result)
        else:
            self.write_with_autoreplace(self.page_former.form_update_downloaded_page())

    def show_logs(self):
        self.write_with_autoreplace(self.page_former.form_show_logs_page())

    def write_log_tail(self):
        logs = log.get_file_tail(log.LOG_FILE)
        content = ''
        for line in logs:
            content = content + line + '<br />'
        if not content:
            content = 'No logs'
        self.write_with_autoreplace(content)

    def ignore_groups_warning(self):
        self.server.app.rights_checker.ignore_flag = True
        time.sleep(self.server.app.rights_checker.CHECK_PERIOD)
        self.do_GET()

    def add_user_groups(self):
        self.server.app.rights_checker_waiter.add_user_groups()
        self.quit_main_app()

    def kill_conveyor(self):
        self.server.app.conveyor_kill_waiter.kill()
        if not self.server.app.conveyor_kill_waiter.waiting:
            message = 'Conveyor was successfully stopped.<br /><br />Returning...'
        else:
            message = '3DPrinterOS was unable to stop MakerBot conveyor service..'
        self.write_message(message)

    def report_problem(self):
        self.write_with_autoreplace(self.page_former.form_report_problem_page())

    def send_problem_report(self):
        summary, description = self.parse_post('problem_summary', 'problem_description')
        report_text = 'Summary:\n' + summary + '\n\nDescription:\n' + description
        error = log.report_problem(report_text)
        if error:
            self.write_message('Reporting problem failed!<br />' + error)
        else:
            self.write_message('Thank you! Your problem was successfully reported.')

    def clear_logs(self):
        log.clear_logs()
        self.write_message('Logs successfully cleared. Writing logs stopped. \
                           <br /><br />Please restart 3DPrinterOS Client to start writing new logs!'
                           + self.page_former.form_back_button('/'), show_time=1)

    def quit_main_app(self):
        self.write_with_autoreplace(self.page_former.form_closing_page())
        self.server.app.stop_flag = True

    def process_login(self):
        login_object = self.server.app.user_login
        if login_object.user_token:
            if login_object.login == self.parse_login()[0]:
                self.answer_on_login(True, 'You are already logged in')
                return
            else:
                error = (0, 'Please logout first before login as other user')
        else:
            login, password, token, force_normal_login = self.parse_login()
            if (login and password) or token:
                error = login_object.login_as_user(login, password, token)
            else:
                error = (1, "Login and password should not be empty")
        if error:
            self.answer_on_login(False, "Login error: " + error[1])
        else:
            if self.command == 'GET' and not force_normal_login:
                #TODO seem to be magic, need to be refactored to clarify
                self.page_former.show_your_account_button = False
                config.get_settings()['camera']['enabled'] = False
                config.get_settings()['web_interface']['browser_opening_on_start'] = False
                self.close_wizard(silent=True)
            self.answer_on_login(True, 'Login successful!<br><br>Processing...')

    def parse_login(self):
        login, password, token = None, None, None
        force_normal_login = False
        if self.command == 'POST':
            login_and_password = self.parse_post('login', 'password')
            if login_and_password and len(login_and_password) == 2:
                login, password = login_and_password
                login = login.replace(' ', '+') # a hack to fix + in email. FIXME right
                password = hashlib.sha256(password.encode('utf-8')).hexdigest()
        elif self.command == 'GET':
            query = urllib.parse.urlparse(self.path).query
            args = urllib.parse.parse_qs(query)
            login = args['user_login'][0]
            if 'auth_token' in args:
                token = args['auth_token'][0]
            elif 'password_hash' in args:
                password = args['password_hash'][0]
            elif 'password' in args:
                password = args['password'][0]
                password = hashlib.sha256(password.encode('utf-8')).hexdigest()
            force_normal_login = "force_normal_login" in args
        return login, password, token, force_normal_login

    def need_image_response(self):
        query = urllib.parse.urlparse(self.path).query
        args = urllib.parse.parse_qs(query)
        return 'image_answer' in args

    def answer_on_login(self, success, message):
        if self.need_image_response():
            if success:
                self.answer_with_image('success.jpg')
            else:
                self.answer_with_image('fail.jpg')
        else:
            self.write_message(message)

    def check_host(self):
        query = urllib.parse.urlparse(self.path).query
        args = urllib.parse.parse_qs(query)
        host_mac = args.get('check_host')[0]
        login_obj = self.server.app.user_login
        if login_obj.user_token:
            if host_mac.lower() == login_obj.mac.lower():
                self.answer_with_image('success.jpg')
                return
        self.answer_with_image('fail.jpg')

    def check_makerware(self):
        if self.server.app.conveyor_kill_waiter.waiting:
            self.answer_with_image('success.jpg')
        else:
            self.answer_with_image('fail.jpg')

    def check_if_printing(self):
        for pi in self.server.app.printer_interfaces:
            if pi.get_printer_state() in ('printing', 'paused', 'downloading'):
                self.answer_with_image('success.jpg')
                return
        self.answer_with_image('fail.jpg')

    def check_linux_rights(self):
        if self.server.app.rights_checker_waiter.waiting:
            self.answer_with_image('success.jpg')
        else:
            self.answer_with_image('fail.jpg')

    def answer_with_image(self, img_filename):
        #this function performs hack that necessary for sending requests from HTTPS to HTTP
        #answering with image is the only way to communicate between HTTPS and HTTP
        message = self.server.web_content_files['static']['images'][img_filename]
        if img_filename.endswith('png'):
            mime = 'image/png'
        elif img_filename.endswith('ico'):
            mime = 'image/ico'
        else:
            mime = 'image/jpeg'
        self.write_with_autoreplace(message, headers = { 'Content-Type': mime })

    def check_login_and_password(self, login, password):
        # check login using RPi's root credentials
        if platforms.get_platform() == 'rpi' and pam:
            try:
                result = pam.authenticate(login, password, service='sudo') #service sudo is essential for this to work on RPi
                if result:
                    return result
            except:
                self.logger.exception("PAM authentication error")
        # check 3dprinteros credentials login 
        password = hashlib.sha256(password.encode('utf-8')).hexdigest()
        stored_login, stored_password = user_login.UserLogin.load_login()
        if stored_login:
            return login == stored_login and password == stored_password

    def create_cookie_token(self):
        #  login = self.server.app.user_login.login
        #  if login:
        #      md5hash = hashlib.md5()
        #      md5hash.update(login.encode("utf-8"))
        #      md5hash.update(self.server.app.user_login.user_token.encode("utf-8"))
        #      md5hash.update(time.strftime('%Y%m%d').encode("utf-8"))
        #      token = md5hash.hexdigest()
        token = secrets.token_hex(self.COOKIE_TOKEN_LENGHT)
        self.server.cookies.append(token)
        return token

    def check_if_rpi(self):
        if platforms.get_platform() == 'rpi':
            return True
        else:
            self.write_message('SSH access configuration available only on Raspberry Pi', show_time = 3)

    def generate_ssh_keys(self):
        if self.check_if_rpi():
            private_key_path = rights.generate_ssh_keys()
            if private_key_path:
                headers = {}
                headers['Content-Type'] = 'application/octet-stream'
                headers['Content-Disposition'] = 'attachment; filename=' + os.path.basename(private_key_path)
                headers['Content-Transfer-Encoding'] = 'binary'
                headers['Accept-Ranges'] = 'bytes'
                with open(private_key_path) as f:
                    key = f.read()
                self.send_response(200)
                self.end_headers()
                for field_name, value in list(headers.items()):
                    self.send_header(field_name, value)
                self.end_headers()
                self.wfile.write(bytes(key, "utf-8"))
                os.remove(private_key_path)
            else:
                message = 'An error occurred during keys generation.'
                self.write_message(message, show_time=5)

    def revoke_all_ssh_keys(self):
        if self.check_if_rpi():
            if rights.revoke_all_ssh_keys():
                message = 'All SSH key are revoked. Now image can only be accessed using password.'
            else:
                message = 'An error occurred during keys revoking.'
            self.write_message(message, show_time=5)

    def add_public_key(self):
        if self.check_if_rpi():
            try:
                public_key = self.parse_post('key')[0]
            except Exception as e:
                message = 'Failed to accept public key file! %s' % str(e)
                self.logger.warning(message)
                self.write_message(message)
            else:
                if rights.add_public_key(public_key):
                    message = 'SSH key was added to authorized keys. You can access image using it.'
                else:
                    message = 'An error occurred during the key addition.'
                self.write_message(message, show_time=5)

    def change_ssh_password(self):
        if self.check_if_rpi():
            password = self.parse_post('password')
            if password is not None:
                if rights.change_ssh_password(password):
                    message = "SSH password for root had been updated."
                else:
                    message = 'An error occurred during password updating.'
                self.write_message(message, show_time=5)

    def telemetry(self):
        app = self.server.app
        telemetry = {}
        telemetry['login'] = getattr(getattr(app, 'user_login', None), 'login', None)
        telemetry['version'] = version.version
        telemetry['update_available'] = app.updater.update_available
        telemetry['camera_type'] = app.camera_controller.current_camera_name
        telemetry['printers'] = [{'usb_info': pi.usb_info, 'report': pi.state_report()} for pi in app.printer_interfaces]
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(json.dumps(telemetry).encode("utf-8"))

    def network_printers_list(self):
        printers_list = self.server.app.network_printers_detector.printers_list_with_profiles()
        printers_list = json.dumps(printers_list)
        self.logger.info("Network printers list:\t" + printers_list)
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(printers_list.encode("utf-8"))

    def forget_network_printer(self):
        usb_info_json = self.parse_post('usb_info')
        self.logger.info("Forgetting network printer: " + str(usb_info_json))
        usb_info = json.loads(usb_info_json)
        self.server.app.network_printers_detector.forget(usb_info)
        self.send_response(200)
        self.end_headers()

    def process_logout(self):
        self.server.app.user_login.logout()
        self.server.app.restart_flag = True
        self.quit_main_app()

    def plugins_list(self):
        plugins_list = list(self.server.app.plugin_controller.plugins.values())
        plugins_list = json.dumps(plugins_list)
        self.logger.info("Printers list:\t" + plugins_list)
        self.send_response(200)
        self.send_header('Content-Type', 'application/json')
        self.end_headers()
        self.wfile.write(plugins_list.encode("utf-8"))

    def control_plugin(self):
        command = self.parse_post('command')
        command = json.loads(command)
        action = command['action']
        method_name = action + "_plugin"
        method = getattr(self.server.app.plugin_controller, method_name, None)
        self.send_response(200)
        self.end_headers()
        if method:
            method(command["plugin_name"])
        else:
            raise RuntimeError("No such plugin action: " + method_name)

    def accept_and_install_plugin_file(self):
        path_to_save_file = os.path.join(paths.current_settings_folder(), "plugin_to_install.zip")
        error = self.parse_and_save_multipart_file_upload(path_to_save_file)
        if error:
            message = 'Failed to accept plugin package! %s' % error
            self.logger.warning(message)
            self.write_message(message)
        else:
            name = self.server.app.plugin_controller.install(path_to_save_file)
            self.server.app.plugin_controller.load_plugin(name)
            os.remove(path_to_save_file)
            self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['plugins_page.html'])

    def print_from_file_or_delete(self):
        filename, printer_id, action = self.parse_post('filename', 'printer_id', 'action')
        filepath = os.path.join(paths.STORAGE_FOLDER, filename)
        self.logger.info("Printer: %s|Action: %s|Filename: %s" % (printer_id, action, filename))
        if action == "Delete":
            try:
                os.remove(filepath)
            except (OSError, IOError):
                message = "Error while removing file %s" % filename
                self.write_message(message)
            else:
                self.local_file_print_or_delete(printer_id)
        elif action == "Print":
            if printer_id and filepath:
                try:
                    printer_interface = self.server.find_printer_interface(printer_id)
                    with open(filepath, "rb") as f:
                        printer_interface.printer.gcodes(f.read())
                except:
                    message = "Error while loading file %s" % filename
                    self.logger.exception(message)
                    self.write_message(message)
                else:
                    self.write_message("File load success. Starting print...")
            else:
                self.write_message("Error: invalid printer ID or file path %s %s" % (printer_id, filepath))

    def upload_to_storage(self):
        error = self.parse_and_save_multipart_file_upload()
        if error:
            message = "Upload failed: %s" % error
        else:
            message = "Upload finished successfully"
        self.write_message(message)

    def upload_to_storage_page(self):
        self.write_with_autoreplace(self.server.web_content_files['templates']['pages']['upload_to_storage.html'])

    def remove_file_from_storage(self):
        self.parse_and_save_multipart_file_upload()

    def local_file_print_or_delete(self, printer_id=None):
        if not printer_id:
            printer_id = self.parse_post("printer_id")
        page = self.server.web_content_files['templates']['pages']['local_file_print_or_delete.html']
        page = page.replace("!!!PRINTER_ID!!!", printer_id)
        page = page.replace("!!!FILES!!!", self.form_storage_files_radio_list())
        page = page.replace("!!!FREE_SPACE!!!", paths.get_human_free_space(paths.STORAGE_FOLDER))
        self.logger.info("Entering local print page for printer:%s" % printer_id)
        self.write_with_autoreplace(page)

    def form_storage_files_radio_list(self):
        radio_template = '<input type="radio" name="filename" value="VALUE">VIEWED_LINE<br />'
        viewed_template = '<b>%s</b>&emsp;<i>%s</i>&emsp;<b>%s</b>'
        output = ''
        for filename in os.listdir(paths.STORAGE_FOLDER):
            filepath = os.path.join(paths.STORAGE_FOLDER, filename)
            access_time = paths.get_human_access_time(filepath)
            size = paths.get_human_file_size(filepath)
            viewed = viewed_template % (filename, access_time, size)
            output += radio_template.replace("VALUE", filename).replace("VIEWED_LINE", viewed)
        return output

    def parse_and_save_multipart_file_upload(self, path=None):
        # NOTE this function will not function properly if there is more than one field in the post
        # NOTE replace with something more reliable - chunks could split boundary and produce a bug
        MULTIPART_READ_CHUNK_SIZE = 10 * 2 ** 20 # 10 MB
        BODY_START_BOUNDARY = b"\r\n\r\n"
        MULTIPART_BOUNDARY_START = b"\r\n----"
        MULTIPART_FILENAME_RE = re.compile(b'filename="(.*)"')
        if not 'multipart/form-data' in self.headers.get('Content-Type', ""):
            message = "Error: call of parse_and_save_multipart_file_upload for non multipart post"
            return message
        bytes_left = int(self.headers.get('Content-Length', 0))
        if not bytes_left:
            message = "Error: no proper Content-Length header"
            return message
        body_start_found = False
        field_end_reached = False
        file_object = None
        self.logger.info("Receiving upload of %d bytes" % bytes_left)
        while bytes_left > 0:
            if self.server.app.stop_flag:
                return "Application quit"
            read_size = min(MULTIPART_READ_CHUNK_SIZE, bytes_left)
            chunk = self.rfile.read(read_size)
            bytes_left -= read_size
            self.logger.info("Received upload chunk: %d bytes" % read_size)
            if not path:
                result = MULTIPART_FILENAME_RE.findall(chunk)
                if result:
                    path = os.path.join(paths.STORAGE_FOLDER, result[0].decode("utf-8"))
            if not body_start_found:
                if BODY_START_BOUNDARY in chunk:
                    chunk = chunk.split(BODY_START_BOUNDARY)[1]
                    body_start_found = True
                    if not path:
                        return "Error finding uploaded file name"
            if body_start_found:
                if not file_object:
                    self.logger.info("Saving uploaded to: %s" % path)
                    file_object = open(path, "wb")
                if not bytes_left: # on last chunk cut other boundary
                    self.logger.info("Upload field ending boundary found")
                    chunk = chunk.split(MULTIPART_BOUNDARY_START)[0]
                    field_end_reached = True
                try:
                    file_object.write(chunk)
                except (IOError, OSError):
                    return "Error writing data to file: %s" % path
                if field_end_reached:
                    return
        if not body_start_found:
            return "Error: post body boundary wasn't found. Cant parse."
        if not field_end_reached:
            return "Error: field end boundary wasn't found"

    def get_detection_wizard_page(self):
        wizard = self.server.app.detection_wizard
        confirm_page_template = self.server.web_content_files['templates']['pages']['confirm.html']
        wizard_template = self.server.web_content_files['templates']['pages']['detection_wizard_page.html']
        if self.server.app.stop_flag:
            wizard.stop()
            self.write_message("Detection canceled by application quit...", show_time=1)
        elif not wizard.active:
            warning = self.htmlify_text(wizard.get_warning_text())
            page = confirm_page_template.replace("!!!BODY!!!", warning)
            page = page.replace("!!!CONFIRM_PATH!!!", "detection_wizard_start")
            page = page.replace("Confirm", "Proceed")
            self.write_with_autoreplace(page)
        else:
            template = '<input type="radio" name="usb_info" value="%s">%s&emsp;<b>%s</b><br />'
            add_locally_button_mark = "!!!ADD_LOCALLY_BUTTON!!!"
            add_locally_button_html = '<input type="submit" name="action" value="Add locally" style="display: inline;\
                    margin-bottom: 0px;" class="big_long_button">'
            devices = wizard.device_status_lines_list_by_id()
            self.logger.debug("Detected devices: %s" % devices)
            lines = [template % device_line for device_line in devices]
            if lines:
                if wizard.scan_finised():
                    wizard_template = wizard_template.replace("Scanning...", "Select a printer to add locally or to add to cloud")
                devices_html = self.htmlify_text("\n".join(lines), ignore_endofline=True)
                wizard_template = wizard_template.replace(add_locally_button_mark, add_locally_button_html)
                self.write_with_autoreplace(wizard_template.replace("!!!DEVICES!!!", devices_html))
            elif wizard.scan_finised():
                devices_html = "Unable to detect any 3D printers"
                wizard_template = wizard_template.replace("Scanning...", "")
                wizard_template = wizard_template.replace(add_locally_button_mark, "")
                self.write_with_autoreplace(wizard_template.replace("!!!DEVICES!!!", devices_html))
            else:
                self.write_message("Scanning for USB devices...", show_time=2, redirect="/detection_wizard_page", button_text="Cancel")

    def detection_wizard_start(self):
        error = self.server.app.detection_wizard.start_scan()
        if error:
            self.write_message(error)
        else:
            self.write_message("Starting Detection Wizard...", show_time=1, redirect="/detection_wizard_page")

    def detection_wizard_action(self):
        detection_wizard = self.server.app.detection_wizard
        results = self.parse_post("action", "usb_info")
        if type(results) in (list, tuple) and len(results) > 1:
            action, usb_info = results
        else:
            action = results
            usb_info = None
        self.logger.debug("Detection wizard action and usb_info: %s %s" % (action, usb_info))
        if action == "Cancel":
            detection_wizard.stop()
            self.write_message("Detection canceled")
        elif not usb_info:
            self.write_message("Please select a printer", show_time=2, redirect="/detection_wizard_page")
        elif action == "Add locally":
            error = detection_wizard.save_profile_locally(usb_info)
            if error:
                self.write_message("Error while saving profile:" + error, show_time=2, redirect="/detection_wizard_page")
            else:
                detection_wizard.stop()
                self.write_message("Printer profiles added. Now you can use them to print locally.")
        elif action == "Go to integration form...":
            self.detection_wizard_integration_form(usb_info)
        else:
            self.write_message("Error: unknown action %s" % action, show_time=2, redirect="/detection_wizard_page")

    def detection_wizard_form_action(self):
        field_names = "model", "manufacturer", "dimensions", "details", "usb_info"
        post_list = self.parse_post(*field_names)
        form_text = ""
        if post_list:
            for index, field in enumerate(field_names):
                if post_list[index]:
                    pretty_field_name = field[0].capitalize() + field[1:]
                    form_text += "%s: %s\n" % (pretty_field_name, str(post_list[index]))
        error = self.server.app.detection_wizard.create_and_send_report(form_text)
        if error:
            message = error
        else:
            message = "Printer integration request was successfully submitted"
        self.write_message(message, show_time=3, redirect="/detection_wizard_page")

    def detection_wizard_integration_form(self, usb_info_str):
        page = self.server.web_content_files['templates']['pages']['detection_wizard_integration_form.html']
        page = page.replace("!!!USB_INFO!!!", usb_info_str)
        try:
            usb_info = json.loads(usb_info_str.replace("'", '"'))
        except json.JSONDecodeError:
            error = "Error decoding usb_info string:%s" % usb_info_str
            self.logger.error(error)
            self.write_message(error, show_time=2, redirect="/detection_wizard_page")
        else:
            dc = self.server.app.detection_wizard.get_device_container_by_usb_info(usb_info)
            text_line = self.server.app.detection_wizard.device_line(dc)
            page = page.replace("!!!TEXT_LINE!!!", text_line)
            self.write_with_autoreplace(page)

    def send_detection_wizard_report(self):
        detection_wizard = self.server.app.detection_wizard
        error = detection_wizard.create_and_send_report()
        if error:
            self.write_message("Error: %s. Please try again." % error, show_time=5, redirect="/detection_wizard_page")
        else:
            detection_wizard.stop()
            self.write_message("Your integration request was successfully submitted", show_time=2, redirect="/detection_wizard_page")
